/*
@license
The following license applies to all parts of this software except as
documented below.

    Copyright (c) 2019, Twilio, inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

      1. Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

      2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in
         the documentation and/or other materials provided with the
         distribution.

      3. Neither the name of Twilio nor the names of its contributors may
         be used to endorse or promote products derived from this software
         without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software includes javascript-state-machine under the following license.

    Copyright (c) 2012, 2013, 2014, 2015, Jake Gordon and contributors

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE

This software includes loglevel under the following license.

    Copyright (c) 2013 Tim Perry

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

This software includes q under the following license.

    Copyright 2009–2014 Kristopher Michael Kowal. All rights reserved.
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.

This software includes platform.js under the following license.

    Copyright 2014 Benjamin Tan <https://d10.github.io/>
    Copyright 2011-2015 John-David Dalton <http://allyoucanleet.com/>

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
var global =
  typeof global !== "undefined"
    ? global
    : typeof self !== "undefined"
    ? self
    : typeof window !== "undefined"
    ? window
    : {};

import { __decorate, __metadata } from './node_modules/tslib/tslib.es6.js';
import { Logger } from './logger.js';
import { Configuration } from './configuration.js';
import { User } from './user.js';
import { Network } from './services/network.js';
import { NotificationTypes } from './interfaces/notification-types.js';
import { InitRegistration, TwilsockClient, TelemetryEventDescription, TelemetryPoint } from 'twilsock';
import { Notifications } from '@twilio/notifications';
import { SyncClient } from 'twilio-sync';
import { CancellablePromise, McsClient } from '@twilio/mcs-client';
import { Conversations } from './data/conversations.js';
import { Users } from './data/users.js';
import { TypingIndicator } from './services/typing-indicator.js';
import { PushNotification } from './push-notification.js';
import { deepClone, parseToNumber } from './util/index.js';
import { validateTypesAsync, nonEmptyString, objectSchema, literal, pureObject, nonEmptyArray, validateTypes, validateConstructorTypes } from '@twilio/declarative-type-validator';
import { version } from './packages/conversations/package.json.js';
import { CommandExecutor } from './command-executor.js';
import { ReplayEventEmitter } from '@twilio/replay-event-emitter';
import { Media } from './media.js';
import { deprecated, deprecationWarning } from '@twilio/deprecation-decorator';
import { ContentClient } from './content-client.js';
import { ChannelMetadataClient } from './channel-metadata-client.js';
import { MessageRecipientsClient } from './message-recipients-client.js';

var Client_1;
/**
 * Client services.
 */
class ClientServices {
}
/**
 * A client is the starting point to the Twilio Conversations functionality.
 */
let Client = Client_1 = class Client extends ReplayEventEmitter {
    /**
     * Returned Conversations Client instance is not yet fully initialized. Calling any
     * operations will block until it is. Use connection events to monitor when
     * client becomes fully available (connectionStateChanged with state
     * 'connected') or not available (connectionStateChange with state 'denied',
     * event tokenExpired, event connectionError).
     *
     * @param fpaToken Access token
     * @param options Options to customize the Client
     * @returns A not yet fully-initialized client.
     */
    constructor(fpaToken, options = {}) {
        var _a, _b, _c, _d, _e;
        super();
        /**
         * Current version of the Conversations client.
         */
        this.version = version;
        /**
         * Client connection state.
         */
        this.connectionState = "unknown";
        /**
         * Parse a push notification payload.
         */
        this.parsePushNotification = Client_1.parsePushNotification;
        this._fpaToken = fpaToken !== null && fpaToken !== void 0 ? fpaToken : "";
        this._options = options !== null && options !== void 0 ? options : {};
        if (!this._options.disableDeepClone) {
            let options = Object.assign(Object.assign({}, this._options), { transport: undefined, twilsockClient: undefined });
            options = deepClone(options);
            options.transport = this._options.transport;
            options.twilsockClient = this._options.twilsockClient;
            this._options = options;
        }
        this._options.logLevel = (_a = this._options.logLevel) !== null && _a !== void 0 ? _a : "silent";
        Client_1._logger.setLevel(this._options.logLevel);
        const productId = (this._options.productId = "ip_messaging");
        // Filling ClientMetadata
        this._options.clientMetadata = this._options.clientMetadata || {};
        if (!this._options.clientMetadata.hasOwnProperty("type")) {
            this._options.clientMetadata.type = "conversations";
        }
        if (!this._options.clientMetadata.hasOwnProperty("sdk")) {
            this._options.clientMetadata.sdk = "JS";
            this._options.clientMetadata.sdkv = version;
        }
        // Enable session local storage for Sync
        this._options.Sync = this._options.Sync || {};
        if (typeof this._options.Sync.enableSessionStorage === "undefined") {
            this._options.Sync.enableSessionStorage = true;
        }
        if (this._options.region) {
            this._options.Sync.region = this._options.region;
        }
        if (!fpaToken) {
            throw new Error("A valid Twilio token should be provided");
        }
        this._services = new ClientServices();
        this._myself = new User("", "", null, this._services);
        const startTwilsock = !this._options.twilsockClient;
        // Create default init registrations if none were provided.
        // Otherwise, the outside party have to list all the init registrations they
        // need.
        // Init registrations passed to the Conversations client will be passed down
        // to the Sync client as well.
        if (!this._options.initRegistrations) {
            const initRegistration = new InitRegistration(productId);
            Client_1.populateInitRegistrations(initRegistration);
            this._options.initRegistrations = [initRegistration];
        }
        this._services.twilsockClient = this._options.twilsockClient =
            (_b = this._options.twilsockClient) !== null && _b !== void 0 ? _b : new TwilsockClient(fpaToken, productId, this._options);
        this._services.twilsockClient.on(Client_1.tokenAboutToExpire, () => this.emit(Client_1.tokenAboutToExpire));
        this._services.twilsockClient.on(Client_1.tokenExpired, () => this.emit(Client_1.tokenExpired));
        this._services.twilsockClient.on(Client_1.connectionError, (error) => this.emit(Client_1.connectionError, error));
        this._services.twilsockClient.on("stateChanged", (state) => {
            Client_1._logger.debug(`Handling stateChanged for ConversationsClient: new state ${state}`);
            if (state !== this.connectionState) {
                this.connectionState = state;
                this.emit(Client_1.connectionStateChanged, this.connectionState);
            }
        });
        this._services.transport = this._options.transport = ((_c = this._options
            .transport) !== null && _c !== void 0 ? _c : this._options.twilsockClient);
        this._services.notificationClient = this._options.notificationsClient =
            (_d = this._options.notificationsClient) !== null && _d !== void 0 ? _d : new Notifications(fpaToken, this._options);
        this._services.syncClient = this._options.syncClient =
            (_e = this._options.syncClient) !== null && _e !== void 0 ? _e : new SyncClient(fpaToken, this._options);
        const configurationOptions = (options === null || options === void 0 ? void 0 : options.Chat) || (options === null || options === void 0 ? void 0 : options.IPMessaging) || options || {};
        const region = configurationOptions.region || (options === null || options === void 0 ? void 0 : options.region);
        const baseUrl = configurationOptions.apiUri ||
            configurationOptions.typingUri ||
            `https://aim.${region || "us1"}.twilio.com`;
        this._services.commandExecutor = new CommandExecutor(baseUrl, { transport: this._options.transport }, productId);
        this._services.contentClient = new ContentClient(this._services);
        const emitFailed = (error) => {
            this._rejectEnsureReady(error);
            this.emit(Client_1.stateChanged, "failed");
            this.emit(Client_1.initFailed, { error });
        };
        const emitDisconnected = () => {
            emitFailed({
                terminal: true,
                message: "Twilsock has disconnected.",
            });
            this._initializeEnsureReady((options === null || options === void 0 ? void 0 : options.throwErrorsAlways) || false);
        };
        this._services.twilsockClient.once("connectionError", emitFailed);
        this._services.twilsockClient.once("disconnected", emitDisconnected);
        this._services.twilsockClient.once("connected", async () => {
            Client_1._logger.debug(`ConversationsClient started INITIALIZING`);
            this._services.twilsockClient.off("connectionError", emitFailed);
            this._services.twilsockClient.off("disconnected", emitDisconnected);
            try {
                const startupEvent = "conversations.client.startup";
                this._services.twilsockClient.addPartialTelemetryEvent(new TelemetryEventDescription(startupEvent, "Conversations client startup", new Date()), startupEvent, TelemetryPoint.Start);
                await this._initialize();
                this._services.twilsockClient.addPartialTelemetryEvent(new TelemetryEventDescription("", "", new Date()), startupEvent, TelemetryPoint.End);
            }
            catch (err) {
                // Fail ChatClient if initialization is incomplete
                const connectionError = {
                    terminal: true,
                    message: err.message,
                };
                this._rejectEnsureReady(connectionError);
                this.emit(Client_1.stateChanged, "failed");
                this.emit(Client_1.initFailed, {
                    error: connectionError,
                });
            }
        });
        this._initializeEnsureReady((options === null || options === void 0 ? void 0 : options.throwErrorsAlways) || false);
        if (startTwilsock) {
            this._services.twilsockClient.connect();
        }
    }
    /**
     * Information of the logged-in user. Before client initialization, returns an
     * uninitialized user. Will trigger a {@link Client.userUpdated} event after
     * initialization.
     */
    get user() {
        return this._myself;
    }
    /**
     * Client reachability state. Throws an error if accessed before the client
     * initialization was completed.
     */
    get reachabilityEnabled() {
        if (!this._configuration) {
            throw new Error("Reachability information could not yet be accessed as the client " +
                "has not yet been initialized. Subscribe to the 'stateChanged' event " +
                "to properly react to the client initialization.");
        }
        return this._configuration.reachabilityEnabled;
    }
    /**
     * @deprecated
     * Current token.
     * @internal
     */
    get token() {
        return this._fpaToken;
    }
    /**
     * @deprecated Call constructor directly.
     *
     * Factory method to create a Conversations client instance.
     *
     * The factory method will automatically trigger connection.
     * Do not use it if you need finer-grained control.
     *
     * Since this method returns an already-initialized client, some of the events
     * will be lost because they happen *before* the initialization. It is
     * recommended that `client.onWithReplay` is used as opposed to `client.on`
     * for subscribing to client events. The `client.onWithReplay` will re-emit
     * the most recent value for a given event if it emitted before the
     * subscription.
     *
     * @param token Access token.
     * @param options Options to customize the client.
     * @returns Returns a fully initialized client.
     */
    static async create(token, options) {
        // The logic is as follows:
        // - If twilsock is not passed in, then the ConversationsClient constructor will call twilsock.connect() by itself
        //   and we do not need to do it here.
        // - If twilsock was passed in from the outside, but customer called ConversationsClient.create() then they are
        //   using an obsolete workflow and the startup sequence will never complete.
        if (options === null || options === void 0 ? void 0 : options.twilsockClient) {
            throw new Error("Obsolete usage of ConversationsClient.create() " +
                "factory method: if you pass twilsock from the outside then you must " +
                "use ConversationsClient constructor and be prepared to work with " +
                "uninitialized client.");
        }
        const client = new Client_1(token, options);
        await client._ensureReady;
        return client;
    }
    /**
     * Static method for push notification payload parsing. Returns parsed push as
     * a {@link PushNotification} object.
     * @param notificationPayload Push notification payload.
     */
    static parsePushNotification(notificationPayload) {
        var _a, _b;
        Client_1._logger.debug("parsePushNotification, notificationPayload=", notificationPayload);
        // APNS specifics
        if (typeof notificationPayload.aps !== "undefined") {
            if (!notificationPayload.twi_message_type) {
                throw new Error("Provided push notification payload does not contain Programmable Chat push notification type");
            }
            const data = Client_1._parsePushNotificationChatData(notificationPayload);
            const apsPayload = notificationPayload.aps;
            let body;
            let title = null;
            if (typeof apsPayload.alert === "string") {
                body = apsPayload.alert || null;
            }
            else {
                body = ((_a = apsPayload.alert) === null || _a === void 0 ? void 0 : _a.body) || null;
                title = ((_b = apsPayload.alert) === null || _b === void 0 ? void 0 : _b.title) || null;
            }
            return new PushNotification({
                title,
                body,
                sound: apsPayload.sound || null,
                badge: apsPayload.badge || null,
                action: apsPayload.category || null,
                type: notificationPayload.twi_message_type,
                data: data,
            });
        }
        // FCM specifics
        if (typeof notificationPayload.data !== "undefined") {
            const dataPayload = notificationPayload.data;
            if (!dataPayload.twi_message_type) {
                throw new Error("Provided push notification payload does not contain Programmable Chat push notification type");
            }
            const data = Client_1._parsePushNotificationChatData(notificationPayload.data);
            return new PushNotification({
                title: dataPayload.twi_title || null,
                body: dataPayload.twi_body || null,
                sound: dataPayload.twi_sound || null,
                badge: null,
                action: dataPayload.twi_action || null,
                type: dataPayload.twi_message_type,
                data: data,
            });
        }
        throw new Error("Provided push notification payload is not Programmable Chat notification");
    }
    /**
     * Static method for parsing push notification chat data.
     * @param data Data to parse
     */
    static _parsePushNotificationChatData(data) {
        const result = {};
        for (const key in Client_1._supportedPushDataFields) {
            const value = data[key];
            if (typeof value === "undefined" || value === null) {
                continue;
            }
            if (key === "message_index" || key === "media_count") {
                const number = parseToNumber(value);
                if (number !== null) {
                    result[Client_1._supportedPushDataFields[key]] = number;
                }
                continue;
            }
            if (key === "media") {
                if (typeof value === "string") {
                    try {
                        result[Client_1._supportedPushDataFields[key]] = JSON.parse(value);
                    }
                    catch (_a) {
                        Client_1._logger.debug("Media message notification parsing error");
                    }
                }
                continue;
            }
            result[Client_1._supportedPushDataFields[key]] = value;
        }
        return result;
    }
    /**
     * Populate the client with init registrations.
     * @param reg The init registration to populate.
     */
    static populateInitRegistrations(reg) {
        reg.populateInitRegistrations([NotificationTypes.TYPING_INDICATOR]);
        SyncClient.populateInitRegistrations(reg);
    }
    /**
     * Gracefully shut down the client.
     */
    async shutdown() {
        await this._ensureReady;
        await this._services.twilsockClient.disconnect();
    }
    /**
     * Update the token used by the client and re-register with the Conversations services.
     * @param token New access token.
     */
    async updateToken(token) {
        await this._ensureReady;
        Client_1._logger.info("updateToken");
        if (this._fpaToken === token) {
            return this;
        }
        await this._services.twilsockClient.updateToken(token);
        await this._services.notificationClient.updateToken(token);
        await this._services.mcsClient.updateToken(token);
        this._fpaToken = token;
        return this;
    }
    /**
     * Get a known conversation by its SID.
     * @param conversationSid Conversation sid
     */
    async getConversationBySid(conversationSid) {
        await this._ensureReady;
        await this._conversationsEntity.myConversationsRead.promise;
        let conversation = await this._conversationsEntity.getConversation(conversationSid);
        if (!conversation) {
            conversation = await this.peekConversationBySid(conversationSid);
            if (conversation) {
                deprecationWarning("The method getConversationBySid is deprecated to retrieve conversations you're not part of. Use peekConversationBySid instead.");
            }
        }
        if (!conversation) {
            throw new Error(`Conversation with SID ${conversationSid} was not found.`);
        }
        return conversation;
    }
    /**
     * Peek a conversation by its SID.
     * @param conversationSid Conversation sid
     * @internal
     */
    async peekConversationBySid(conversationSid) {
        await this._ensureReady;
        const conversation = await this._conversationsEntity.peekConversation(conversationSid);
        if (!conversation) {
            throw new Error(`Conversation with SID ${conversationSid} was not found.`);
        }
        return conversation;
    }
    /**
     * Get a known conversation by its unique identifier name.
     * @param uniqueName The unique identifier name of the conversation.
     */
    async getConversationByUniqueName(uniqueName) {
        await this._ensureReady;
        await this._conversationsEntity.myConversationsRead.promise;
        const conversation = await this._conversationsEntity.getConversationByUniqueName(uniqueName);
        if (!conversation) {
            throw new Error(`Conversation with unique name ${uniqueName} was not found.`);
        }
        return conversation;
    }
    /**
     * Get the current list of all the subscribed conversations.
     */
    async getSubscribedConversations() {
        await this._ensureReady;
        return this._conversationsPromise.then((conversations) => conversations.getConversations());
    }
    /**
     * Create a conversation on the server and subscribe to its events.
     * The default is a conversation with an empty friendly name.
     * @param options Options for the conversation.
     */
    async createConversation(options) {
        await this._ensureReady;
        options = options || {};
        return this._conversationsPromise.then((conversationsEntity) => conversationsEntity.addConversation(options));
    }
    /**
     * Register for push notifications.
     * @param channelType Channel type.
     * @param registrationId Push notification ID provided by the FCM/APNS service
     * on the platform.
     */
    async setPushRegistrationId(channelType, registrationId) {
        await this._ensureReady;
        this._subscribeToPushNotifications(channelType);
        this._services.notificationClient.setPushRegistrationId(channelType, registrationId);
        await this._services.notificationClient.commitChanges(); // Committing before this point is useless because we have no push id
    }
    /**
     * Unregister from push notifications.
     * @param channelType Channel type.
     * @deprecated Use removePushRegistrations() instead.
     */
    async unsetPushRegistrationId(channelType) {
        await this._ensureReady;
        this._unsubscribeFromPushNotifications(channelType);
        await this._services.notificationClient.commitChanges();
    }
    /**
     * Clear existing registrations directly using provided device token.
     * This is useful to ensure stopped subscriptions without resubscribing.
     *
     * This function goes completely beside the state machine and removes all
     * registrations.
     * Use with caution: if it races with current state machine operations,
     * madness will ensue.
     *
     * @param channelType Channel type.
     * @param registrationId Push notification ID provided by the FCM/APNS service
     * on the platform.
     */
    async removePushRegistrations(channelType, registrationId) {
        // do not await this._ensureReady() here - it could be called at any moment
        await this._services.notificationClient.removeRegistrations(channelType, registrationId);
    }
    /**
     * Handle push notification payload parsing and emit the
     * {@link Client.pushNotification} event on this {@link Client} instance.
     * @param notificationPayload Push notification payload
     */
    async handlePushNotification(notificationPayload) {
        await this._ensureReady;
        Client_1._logger.debug("handlePushNotification, notificationPayload=", notificationPayload);
        this.emit("pushNotification", Client_1.parsePushNotification(notificationPayload));
    }
    /**
     * Gets a user with the given identity. If it's in the subscribed list, then
     * return the user object from it;
     * if not, then subscribe and add user to the subscribed list.
     * @param identity Identity of the user.
     * @returns A fully initialized user.
     */
    async getUser(identity) {
        await this._ensureReady;
        return this._services.users.getUser(identity);
    }
    /**
     * Get a list of subscribed user objects.
     */
    async getSubscribedUsers() {
        await this._ensureReady;
        return this._services.users.getSubscribedUsers();
    }
    /**
     * Get content URLs for all media attachments in the given set of media sids
     * using a single operation.
     * @param mediaSids Set of media sids to query for the content URL.
     */
    getTemporaryContentUrlsForMediaSids(mediaSids) {
        return new CancellablePromise(async (resolve, reject, onCancel) => {
            if (!this._services.mcsClient || !mediaSids) {
                reject(new Error("Media Content Service is unavailable"));
                return;
            }
            const request = this._services.mcsClient.mediaSetGetContentUrls(mediaSids);
            onCancel(() => {
                request.cancel();
            });
            try {
                const urls = await request;
                resolve(urls);
            }
            catch (e) {
                reject(e);
            }
        });
    }
    /**
     * Get content URLs for all media attachments in the given set using a single
     * operation.
     * @param contentSet Set of media attachments to query content URLs.
     */
    getTemporaryContentUrlsForMedia(contentSet) {
        // We ignore existing mcsMedia members of each of the media entries.
        // Instead, we just collect their sids and pull new descriptors from a
        // mediaSet GET endpoint.
        const sids = contentSet.map((m) => m.sid);
        return this.getTemporaryContentUrlsForMediaSids(sids);
    }
    /**
     * Returns rich content templates belonging to the account. Rich content
     * templates can be created via the Twilio console or the REST API.
     */
    async getContentTemplates() {
        await this._ensureReady;
        return await this._services.contentClient.getContentTemplates();
    }
    /**
     * Initialize the client.
     */
    async _initialize() {
        const configurationResponse = await this._services.commandExecutor.fetchResource("Client/v2/Configuration");
        this._configuration = new Configuration(this._options, configurationResponse, Client_1._logger);
        this._services.channelMetadataClient = new ChannelMetadataClient(this._services, this._configuration);
        this._services.messageRecipientsClient = new MessageRecipientsClient(this._services, this._configuration);
        this._myself._resolveInitialization(this._configuration, this._configuration.userIdentity, this._configuration.userInfo, true);
        this._services.typingIndicator = new TypingIndicator(this.getConversationBySid.bind(this), this._configuration, this._services);
        this._services.network = new Network(this._configuration, this._services);
        this._services.users = new Users(this._myself, this._configuration, this._services);
        this._services.users.on("userSubscribed", (user) => {
            this.emit("userSubscribed", user);
        });
        this._services.users.on("userUpdated", (args) => this.emit("userUpdated", args));
        this._services.users.on("userUnsubscribed", (user) => {
            this.emit("userUnsubscribed", user);
        });
        this._conversationsEntity = new Conversations(this._configuration, this._services);
        this._conversationsEntity.on("conversationAdded", (conversation) => {
            this.emit("conversationAdded", conversation);
        });
        this._conversationsEntity.on("conversationRemoved", (conversation) => {
            this.emit("conversationRemoved", conversation);
        });
        this._conversationsEntity.on("conversationJoined", (conversation) => {
            this.emit("conversationJoined", conversation);
        });
        this._conversationsEntity.on("conversationLeft", (conversation) => {
            this.emit("conversationLeft", conversation);
        });
        this._conversationsEntity.on("conversationUpdated", (args) => this.emit("conversationUpdated", args));
        this._conversationsEntity.on("participantJoined", (participant) => {
            this.emit("participantJoined", participant);
        });
        this._conversationsEntity.on("participantLeft", (participant) => {
            this.emit("participantLeft", participant);
        });
        this._conversationsEntity.on("participantUpdated", (args) => this.emit("participantUpdated", args));
        this._conversationsEntity.on("messageAdded", (message) => this.emit("messageAdded", message));
        this._conversationsEntity.on("messageUpdated", (args) => this.emit("messageUpdated", args));
        this._conversationsEntity.on("messageRemoved", (message) => this.emit("messageRemoved", message));
        this._conversationsEntity.on("typingStarted", (participant) => this.emit("typingStarted", participant));
        this._conversationsEntity.on("typingEnded", (participant) => this.emit("typingEnded", participant));
        this._conversationsPromise = this._conversationsEntity
            .fetchConversations()
            .then(() => this._conversationsEntity)
            .catch((error) => {
            console.error("Failed to fetch conversations _conversationsPromise -> client.ts", error);
            throw error;
        });
        await this._services.users.myself._ensureFetched();
        Client_1._supportedPushChannels.forEach((channelType) => this._subscribeToPushNotifications(channelType));
        this._services.typingIndicator.initialize();
        this._services.mcsClient = new McsClient(this._fpaToken, this._configuration.links.mediaService, this._configuration.links.mediaSetService, Object.assign(Object.assign({}, this._options), { transport: undefined }));
        this._resolveEnsureReady();
        this.emit(Client_1.stateChanged, "initialized");
        this.emit(Client_1.initialized);
    }
    /**
     * Subscribe to push notifications.
     * @param channelType The channel type to subscribe to.
     */
    _subscribeToPushNotifications(channelType) {
        [
            NotificationTypes.NEW_MESSAGE,
            NotificationTypes.ADDED_TO_CONVERSATION,
            NotificationTypes.REMOVED_FROM_CONVERSATION,
            NotificationTypes.TYPING_INDICATOR,
            NotificationTypes.CONSUMPTION_UPDATE,
        ].forEach((messageType) => {
            this._services.notificationClient.subscribe(channelType, messageType);
        });
    }
    /**
     * Unsubscribe from push notifications.
     * @param channelType The channel type to unsubscribe from.
     */
    _unsubscribeFromPushNotifications(channelType) {
        [
            NotificationTypes.NEW_MESSAGE,
            NotificationTypes.ADDED_TO_CONVERSATION,
            NotificationTypes.REMOVED_FROM_CONVERSATION,
            NotificationTypes.TYPING_INDICATOR,
            NotificationTypes.CONSUMPTION_UPDATE,
        ].forEach((messageType) => {
            this._services.notificationClient.unsubscribe(channelType, messageType);
        });
    }
    /**
     * Initialize the ensureReady promise.
     */
    _initializeEnsureReady(throwErrorsAlways) {
        this._ensureReady = new Promise((resolve, reject) => {
            this._resolveEnsureReady = resolve;
            this._rejectEnsureReady = reject;
        }).catch((error) => {
            if (throwErrorsAlways) {
                throw error;
            }
            else {
                return void 0;
            }
        });
    }
};
/**
 * Fired when a conversation becomes visible to the client. The event is also
 * triggered when the client creates a new conversation.
 * Fired for all conversations that the client has joined.
 *
 * Parameters:
 * 1. {@link Conversation} `conversation` - the conversation in question
 * @event
 */
Client.conversationAdded = "conversationAdded";
/**
 * Fired when the client joins a conversation.
 *
 * Parameters:
 * 1. {@link Conversation} `conversation` - the conversation in question
 * @event
 */
Client.conversationJoined = "conversationJoined";
/**
 * Fired when the client leaves a conversation.
 *
 * Parameters:
 * 1. {@link Conversation} `conversation` - the conversation in question
 * @event
 */
Client.conversationLeft = "conversationLeft";
/**
 * Fired when a conversation is no longer visible to the client.
 *
 * Parameters:
 * 1. {@link Conversation} `conversation` - the conversation in question
 * @event
 */
Client.conversationRemoved = "conversationRemoved";
/**
 * Fired when the attributes or the metadata of a conversation have been
 * updated. During conversation's creation and initialization, this event
 * might be fired multiple times for same joined or created conversation as
 * new data is arriving from different sources.
 *
 * Parameters:
 * 1. object `data` - info object provided with the event. It has the
 * following properties:
 *     * {@link Conversation} `conversation` - the conversation in question
 *     * {@link ConversationUpdateReason}[] `updateReasons` - array of reasons
 *     for the update
 * @event
 */
Client.conversationUpdated = "conversationUpdated";
/**
 * Fired when a participant has joined a conversation.
 *
 * Parameters:
 * 1. {@link Participant} `participant` - the participant in question
 * @event
 */
Client.participantJoined = "participantJoined";
/**
 * Fired when a participant has left a conversation.
 *
 * Parameters:
 * 1. {@link Participant} `participant` - the participant in question
 * @event
 */
Client.participantLeft = "participantLeft";
/**
 * Fired when a participant's fields have been updated.
 *
 * Parameters:
 * 1. object `data` - info object provided with the event. It has the
 * following properties:
 *     * {@link Participant} `participant` - the participant in question
 *     * {@link ParticipantUpdateReason}[] `updateReasons` - array of reasons
 *     for the update
 * @event
 */
Client.participantUpdated = "participantUpdated";
/**
 * Fired when a new message has been added to the conversation on the server.
 *
 * Parameters:
 * 1. {@link Message} `message` - the message in question
 * @event
 */
Client.messageAdded = "messageAdded";
/**
 * Fired when a message is removed from the message list of a conversation.
 *
 * Parameters:
 * 1. {@link Message} `message` - the message in question
 * @event
 */
Client.messageRemoved = "messageRemoved";
/**
 * Fired when the fields of an existing message are updated with new values.
 *
 * Parameters:
 * 1. object `data` - info object provided with the event. It has the
 * following properties:
 *     * {@link Message} `message` - the message in question
 *     * {@link MessageUpdateReason}[] `updateReasons` - array of reasons for
 *     the update
 * @event
 */
Client.messageUpdated = "messageUpdated";
/**
 * Fired when the token is about to expire and needs to be updated.
 * @event
 */
Client.tokenAboutToExpire = "tokenAboutToExpire";
/**
 * Fired when the token has expired.
 * @event
 */
Client.tokenExpired = "tokenExpired";
/**
 * Fired when a participant has stopped typing.
 *
 * Parameters:
 * 1. {@link Participant} `participant` - the participant in question
 * @event
 */
Client.typingEnded = "typingEnded";
/**
 * Fired when a participant has started typing.
 *
 * Parameters:
 * 1. {@link Participant} `participant` - the participant in question
 * @event
 */
Client.typingStarted = "typingStarted";
/**
 * Fired when the client has received (and parsed) a push notification via one
 * of the push channels (apn or fcm).
 *
 * Parameters:
 * 1. {@link PushNotification} `pushNotification` - the push notification in
 * question
 * @event
 */
Client.pushNotification = "pushNotification";
/**
 * Fired when the client has subscribed to a user.
 *
 * Parameters:
 * 1. {@link User} `user` - the user in question
 * @event
 */
Client.userSubscribed = "userSubscribed";
/**
 * Fired when the client has unsubscribed from a user.
 *
 * Parameters:
 * 1. {@link User} `user` - the user in question
 * @event
 */
Client.userUnsubscribed = "userUnsubscribed";
/**
 * Fired when the properties or the reachability status of a user have been
 * updated.
 *
 * Parameters:
 * 1. object `data` - info object provided with the event. It has the
 * following properties:
 *     * {@link User} `user` - the user in question
 *     * {@link UserUpdateReason}[] `updateReasons` - array of reasons for the
 *     update
 * @event
 */
Client.userUpdated = "userUpdated";
/**
 * @deprecated Use initialized or initFailed events instead
 * Fired when the state of the client has been changed.
 *
 * Parameters:
 * 1. {@link State} `state` - the new client state
 * @event
 */
Client.stateChanged = "stateChanged";
/**
 * Fired when the client has completed initialization successfully.
 * @event
 */
Client.initialized = "initialized";
/**
 * Fired when the client initialization failed.
 *
 * Parameters:
 * 1. object `data` - info object provided with the event. It has the
 * following property:
 *     * Error? `error` - the initialization error if present
 * @event
 */
Client.initFailed = "initFailed";
/**
 * Fired when the connection state of the client has been changed.
 *
 * Parameters:
 * 1. {@link ConnectionState} `state` - the new connection state
 * @event
 */
Client.connectionStateChanged = "connectionStateChanged";
/**
 * Fired when the connection is interrupted for an unexpected reason.
 *
 * Parameters:
 * 1. object `data` - info object provided with the event. It has the
 * following properties:
 *     * boolean `terminal` - Twilsock will stop connection attempts if true
 *     * string `message` - the error message of the root cause
 *     * number? `httpStatusCode` - http status code if available
 *     * number? `errorCode` - Twilio public error code if available
 * @event
 */
Client.connectionError = "connectionError";
/**
 * Current version of the Conversations client.
 */
Client.version = version;
/**
 * Logger instance.
 */
Client._logger = Logger.scope("Client");
/**
 * Supported push notification channels.
 */
Client._supportedPushChannels = [
    "fcm",
    "apn",
];
/**
 * Supported push data fields.
 */
Client._supportedPushDataFields = {
    conversation_sid: "conversationSid",
    conversation_title: "conversationTitle",
    message_sid: "messageSid",
    message_index: "messageIndex",
    media_count: "mediaCount",
    media: "media", // object
};
__decorate([
    deprecated("token"),
    __metadata("design:type", String),
    __metadata("design:paramtypes", [])
], Client.prototype, "token", null);
__decorate([
    validateTypesAsync(nonEmptyString),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], Client.prototype, "updateToken", null);
__decorate([
    validateTypesAsync(nonEmptyString),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], Client.prototype, "getConversationBySid", null);
__decorate([
    validateTypesAsync(nonEmptyString),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], Client.prototype, "peekConversationBySid", null);
__decorate([
    validateTypesAsync(nonEmptyString),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], Client.prototype, "getConversationByUniqueName", null);
__decorate([
    validateTypesAsync([
        "undefined",
        objectSchema("conversation options", {
            friendlyName: ["string", "undefined"],
            isPrivate: ["boolean", "undefined"],
            uniqueName: ["string", "undefined"],
        }),
    ]),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Promise)
], Client.prototype, "createConversation", null);
__decorate([
    validateTypesAsync(literal("fcm", "apn"), "string"),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, String]),
    __metadata("design:returntype", Promise)
], Client.prototype, "setPushRegistrationId", null);
__decorate([
    validateTypesAsync(literal("fcm", "apn")),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], Client.prototype, "unsetPushRegistrationId", null);
__decorate([
    validateTypesAsync(literal("fcm", "apn"), nonEmptyString),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, String]),
    __metadata("design:returntype", Promise)
], Client.prototype, "removePushRegistrations", null);
__decorate([
    validateTypesAsync(pureObject),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Promise)
], Client.prototype, "handlePushNotification", null);
__decorate([
    validateTypesAsync(nonEmptyString),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], Client.prototype, "getUser", null);
__decorate([
    validateTypesAsync(nonEmptyArray("strings", "string")),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", CancellablePromise)
], Client.prototype, "getTemporaryContentUrlsForMediaSids", null);
__decorate([
    validateTypesAsync(nonEmptyArray("media", Media)),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", CancellablePromise)
], Client.prototype, "getTemporaryContentUrlsForMedia", null);
__decorate([
    deprecated("Client.create()", "new Client()"),
    validateTypesAsync("string", ["undefined", pureObject]),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String, Object]),
    __metadata("design:returntype", Promise)
], Client, "create", null);
__decorate([
    validateTypes(pureObject),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", PushNotification)
], Client, "parsePushNotification", null);
Client = Client_1 = __decorate([
    validateConstructorTypes(nonEmptyString, [pureObject, "undefined"]),
    __metadata("design:paramtypes", [String, Object])
], Client);

export { Client };
//# sourceMappingURL=client.js.map
