/*
@license
The following license applies to all parts of this software except as
documented below.

    Copyright (c) 2019, Twilio, inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

      1. Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

      2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in
         the documentation and/or other materials provided with the
         distribution.

      3. Neither the name of Twilio nor the names of its contributors may
         be used to endorse or promote products derived from this software
         without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software includes javascript-state-machine under the following license.

    Copyright (c) 2012, 2013, 2014, 2015, Jake Gordon and contributors

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE

This software includes loglevel under the following license.

    Copyright (c) 2013 Tim Perry

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

This software includes q under the following license.

    Copyright 2009–2014 Kristopher Michael Kowal. All rights reserved.
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.

This software includes platform.js under the following license.

    Copyright 2014 Benjamin Tan <https://d10.github.io/>
    Copyright 2011-2015 John-David Dalton <http://allyoucanleet.com/>

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
var global =
  typeof global !== "undefined"
    ? global
    : typeof self !== "undefined"
    ? self
    : typeof window !== "undefined"
    ? window
    : {};

import { Participant } from '../participant.js';
import { Logger } from '../logger.js';
import { Conversation } from '../conversation.js';
import { SyncMap } from 'twilio-sync';
import { ReplayEventEmitter } from '@twilio/replay-event-emitter';

const log = Logger.scope("Participants");
/**
 * @classdesc Represents the collection of participants for the conversation
 * @fires Participants#participantJoined
 * @fires Participants#participantLeft
 * @fires Participants#participantUpdated
 */
class Participants extends ReplayEventEmitter {
    constructor(conversation, participants, links, services) {
        super();
        this.rosterEntityPromise = null;
        this.conversation = conversation;
        this.participants = participants;
        this.links = links;
        this.services = services;
    }
    async unsubscribe() {
        if (this.rosterEntityPromise) {
            const entity = await this.rosterEntityPromise;
            entity.close();
            this.rosterEntityPromise = null;
        }
    }
    subscribe(arg) {
        const participantsMapPromise = typeof arg === "string"
            ? this.services.syncClient.map({ id: arg, mode: "open_existing" })
            : Promise.resolve(arg);
        return (this.rosterEntityPromise =
            this.rosterEntityPromise ||
                participantsMapPromise
                    .then((rosterMap) => {
                    rosterMap.on(SyncMap.itemAdded, (args) => {
                        log.debug(this.conversation.sid + " itemAdded: " + args.item.key);
                        this.upsertParticipant(args.item.key, args.item.data).then((participant) => {
                            this.emit(Conversation.participantJoined, participant);
                        });
                    });
                    rosterMap.on(SyncMap.itemRemoved, (args) => {
                        log.debug(this.conversation.sid + " itemRemoved: " + args.key);
                        const participantSid = args.key;
                        if (!this.participants.has(participantSid)) {
                            return;
                        }
                        const leftParticipant = this.participants.get(participantSid);
                        this.participants.delete(participantSid);
                        if (!leftParticipant) {
                            return;
                        }
                        this.emit(Conversation.participantLeft, leftParticipant);
                    });
                    rosterMap.on(SyncMap.itemUpdated, (args) => {
                        log.debug(this.conversation.sid + " itemUpdated: " + args.item.key);
                        this.upsertParticipant(args.item.key, args.item.data).catch((e) => log.error(e));
                    });
                    const participantsPromises = [];
                    const rosterMapHandler = (paginator) => {
                        paginator.items.forEach((item) => {
                            participantsPromises.push(this.upsertParticipant(item.key, item.data));
                        });
                        return paginator.hasNextPage
                            ? paginator.nextPage().then(rosterMapHandler)
                            : null;
                    };
                    return rosterMap
                        .getItems()
                        .then(rosterMapHandler)
                        .then(() => Promise.all(participantsPromises))
                        .then(() => rosterMap);
                })
                    .catch((err) => {
                    this.rosterEntityPromise = null;
                    if (this.services.syncClient.connectionState != "disconnected") {
                        log.error("Failed to get roster object for conversation", this.conversation.sid, err);
                    }
                    log.debug("ERROR: Failed to get roster object for conversation", this.conversation.sid, err);
                    throw err;
                }));
    }
    async upsertParticipantFromResponse(data) {
        var _a;
        const { sid, attributes: responseAttributes, date_created: dateCreated, date_updated: dateUpdated, identity: responseIdentity, role_sid: roleSid, messaging_binding: messagingBinding, } = data;
        return await this.upsertParticipant(sid, {
            attributes: responseAttributes,
            dateCreated: new Date(dateCreated),
            dateUpdated: new Date(dateUpdated),
            identity: responseIdentity,
            roleSid,
            lastConsumedMessageIndex: null,
            lastConsumptionTimestamp: null,
            type: (_a = messagingBinding === null || messagingBinding === void 0 ? void 0 : messagingBinding.type) !== null && _a !== void 0 ? _a : "chat",
        });
    }
    async upsertParticipant(participantSid, data) {
        let participant = this.participants.get(participantSid);
        if (participant) {
            return participant._update(data);
        }
        const links = {
            self: `${this.links.participants}/${participantSid}`,
        };
        participant = new Participant(data, participantSid, this.conversation, links, this.services);
        this.participants.set(participantSid, participant);
        participant.on(Participant.updated, (args) => this.emit(Conversation.participantUpdated, args));
        return participant;
    }
    /**
     * @returns {Promise<Array<Participant>>} returns list of participants {@see Participant}
     */
    async getParticipants() {
        return this.rosterEntityPromise
            ? this.rosterEntityPromise.then(() => {
                const participants = [];
                this.participants.forEach((participant) => participants.push(participant));
                return participants;
            })
            : [];
    }
    /**
     * Get participant by SID from conversation
     * @returns {Promise<Participant>}
     */
    async getParticipantBySid(participantSid) {
        return this.rosterEntityPromise
            ? this.rosterEntityPromise.then(() => {
                const participant = this.participants.get(participantSid);
                if (!participant) {
                    throw new Error("Participant with SID " + participantSid + " was not found");
                }
                return participant;
            })
            : null;
    }
    /**
     * Get participant by identity from conversation
     * @returns {Promise<Participant>}
     */
    async getParticipantByIdentity(identity) {
        let foundParticipant = null;
        return this.rosterEntityPromise
            ? this.rosterEntityPromise.then(() => {
                this.participants.forEach((participant) => {
                    if (participant.identity === identity) {
                        foundParticipant = participant;
                    }
                });
                if (!foundParticipant) {
                    throw new Error("Participant with identity " + identity + " was not found");
                }
                return foundParticipant;
            })
            : null;
    }
    /**
     * Add a chat participant to the conversation
     */
    async add(identity, attributes) {
        return await this.services.commandExecutor.mutateResource("post", this.links.participants, {
            identity,
            attributes: typeof attributes !== "undefined"
                ? JSON.stringify(attributes)
                : undefined,
        });
    }
    /**
     * Add a non-chat participant to the conversation.
     */
    async addNonChatParticipant(proxyAddress, address, attributes = {}, bindingOptions = {}) {
        var _a, _b;
        return await this.services.commandExecutor.mutateResource("post", this.links.participants, {
            attributes: typeof attributes !== "undefined"
                ? JSON.stringify(attributes)
                : undefined,
            messaging_binding: {
                address,
                proxy_address: proxyAddress,
                name: (_a = bindingOptions === null || bindingOptions === void 0 ? void 0 : bindingOptions.email) === null || _a === void 0 ? void 0 : _a.name,
                level: (_b = bindingOptions === null || bindingOptions === void 0 ? void 0 : bindingOptions.email) === null || _b === void 0 ? void 0 : _b.level,
            },
        });
    }
    /**
     * Remove the participant with a given identity from a conversation.
     */
    remove(identity) {
        return this.services.commandExecutor.mutateResource("delete", `${this.links.participants}/${encodeURIComponent(identity)}`);
    }
}
/**
 * Fired when participant joined conversation
 * @event Participants#participantJoined
 * @type {Participant}
 */
/**
 * Fired when participant left conversation
 * @event Participants#participantLeft
 * @type {Participant}
 */
/**
 * Fired when participant updated
 * @event Participants#participantUpdated
 * @type {Object}
 * @property {Participant} participant - Updated Participant
 * @property {Participant#UpdateReason[]} updateReasons - Array of Participant's updated event reasons
 */

export { Participants };
//# sourceMappingURL=participants.js.map
