/*
@license
The following license applies to all parts of this software except as
documented below.

    Copyright (c) 2019, Twilio, inc.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

      1. Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.

      2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in
         the documentation and/or other materials provided with the
         distribution.

      3. Neither the name of Twilio nor the names of its contributors may
         be used to endorse or promote products derived from this software
         without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software includes javascript-state-machine under the following license.

    Copyright (c) 2012, 2013, 2014, 2015, Jake Gordon and contributors

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE

This software includes loglevel under the following license.

    Copyright (c) 2013 Tim Perry

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

This software includes q under the following license.

    Copyright 2009–2014 Kristopher Michael Kowal. All rights reserved.
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.

This software includes platform.js under the following license.

    Copyright 2014 Benjamin Tan <https://d10.github.io/>
    Copyright 2011-2015 John-David Dalton <http://allyoucanleet.com/>

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
var global =
  typeof global !== "undefined"
    ? global
    : typeof self !== "undefined"
    ? self
    : typeof window !== "undefined"
    ? window
    : {};

import { __decorate, __metadata, __classPrivateFieldGet } from './node_modules/tslib/tslib.es6.js';
import { parseAttributes, UriBuilder } from './util/index.js';
import { Logger } from './logger.js';
import { CancellablePromise } from '@twilio/mcs-client';
import { Media } from './media.js';
import { AggregatedDeliveryReceipt } from './aggregated-delivery-receipt.js';
import { validateTypes, nonEmptyString, validateTypesAsync, nonEmptyArray } from '@twilio/declarative-type-validator';
import { json } from './interfaces/rules.js';
import { RestPaginator } from './rest-paginator.js';
import { DetailedDeliveryReceipt } from './detailed-delivery-receipt.js';
import { ReplayEventEmitter } from '@twilio/replay-event-emitter';
import isEqual from 'lodash.isequal';
import { deprecated } from '@twilio/deprecation-decorator';
import { parseVariant } from './content-template.js';

var _Message_instances, _Message_wrapMedia, _Message_assignMedias;
const log = Logger.scope("Message");
const XHR = 
// eslint-disable-next-line @typescript-eslint/no-var-requires
global["XMLHttpRequest"] || {};
/**
 * A message in a conversation.
 */
class Message extends ReplayEventEmitter {
    /**
     * @internal
     */
    constructor(index, data, conversation, links, configuration, services) {
        var _a, _b, _c, _d, _e, _f, _g;
        super();
        _Message_instances.add(this);
        this.conversation = conversation;
        this.links = links;
        this.configuration = configuration;
        this.services = services;
        this.state = {
            sid: data.sid,
            index: index,
            author: data.author,
            subject: data.subject,
            contentSid: data.contentSid,
            body: (_a = data.text) !== null && _a !== void 0 ? _a : null,
            timestamp: data.timestamp ? new Date(data.timestamp) : null,
            dateUpdated: data.dateUpdated ? new Date(data.dateUpdated) : null,
            lastUpdatedBy: (_b = data.lastUpdatedBy) !== null && _b !== void 0 ? _b : null,
            attributes: parseAttributes(data.attributes, `Got malformed attributes for the message ${data.sid}`, log),
            type: (_c = data.type) !== null && _c !== void 0 ? _c : "text",
            media: null,
            medias: null,
            participantSid: (_d = data.memberSid) !== null && _d !== void 0 ? _d : null,
            aggregatedDeliveryReceipt: data.delivery
                ? new AggregatedDeliveryReceipt(data.delivery)
                : null,
            hasChannelMetadata: (_e = data.channelMetadata) !== null && _e !== void 0 ? _e : false,
        };
        __classPrivateFieldGet(this, _Message_instances, "m", _Message_assignMedias).call(this, (_f = data.medias) !== null && _f !== void 0 ? _f : null, (_g = data.media) !== null && _g !== void 0 ? _g : null);
    }
    /**
     * The server-assigned unique identifier for the message.
     */
    get sid() {
        return this.state.sid;
    }
    /**
     * Name of the user that sent the message.
     */
    get author() {
        return this.state.author;
    }
    /**
     * Message subject. Used only in email conversations.
     */
    get subject() {
        return this.state.subject;
    }
    /**
     * Unique identifier of {@link ContentTemplate} for this message.
     */
    get contentSid() {
        return this.state.contentSid;
    }
    /**
     * Body of the message.
     */
    get body() {
        return this.state.body;
    }
    /**
     * Date this message was last updated on.
     */
    get dateUpdated() {
        return this.state.dateUpdated;
    }
    /**
     * Index of this message in the conversation's list of messages.
     *
     * By design, the message indices may have arbitrary gaps between them,
     * that does not necessarily mean they were deleted or otherwise modified - just that
     * messages may have some non-contiguous indices even if they are being sent immediately one after another.
     *
     * Trying to use indices for some calculations is going to be unreliable.
     *
     * To calculate the number of unread messages, it is better to use the Read Horizon API.
     * See {@link Conversation.getUnreadMessagesCount} for details.
     */
    get index() {
        return this.state.index;
    }
    /**
     * Identity of the last user that updated the message.
     */
    get lastUpdatedBy() {
        return this.state.lastUpdatedBy;
    }
    /**
     * Date this message was created on.
     */
    get dateCreated() {
        return this.state.timestamp;
    }
    /**
     * Custom attributes of the message.
     */
    get attributes() {
        return this.state.attributes;
    }
    /**
     * Type of the message.
     */
    get type() {
        return this.state.type;
    }
    /**
     * One of the attached media (if present).
     * @deprecated Use attachedMedia instead. Note that the latter is now an array.
     */
    get media() {
        return this.state.media;
    }
    /**
     * Return all media attachments, except email body/history attachments, without temporary urls.
     */
    get attachedMedia() {
        return this.getMediaByCategories(["media"]);
    }
    /**
     * The server-assigned unique identifier of the authoring participant.
     */
    get participantSid() {
        return this.state.participantSid;
    }
    /**
     * Aggregated information about the message delivery statuses across all participants of a conversation..
     */
    get aggregatedDeliveryReceipt() {
        return this.state.aggregatedDeliveryReceipt;
    }
    /**
     * @deprecated
     * Return a (possibly empty) array of media matching a specific set of categories.
     * Allowed category is so far only 'media'.
     * @param categories Array of categories to match.
     * @returns Array of media descriptors matching given categories.
     */
    getMediaByCategory(categories) {
        return this.getMediaByCategories(categories);
    }
    /**
     * Return a (possibly empty) array of media matching a specific set of categories.
     * Allowed category is so far only 'media'.
     * @param categories Array of categories to match.
     * @returns Array of media descriptors matching given categories.
     */
    getMediaByCategories(categories) {
        var _a;
        return ((_a = this.state.medias) !== null && _a !== void 0 ? _a : []).filter((m) => categories.includes(m.category));
    }
    /**
     * Get a media descriptor for an email body attachment of a provided type.
     * Allowed body types are returned in the Conversation.limits().emailBodiesAllowedContentTypes array.
     * @param type Type of email body to request, defaults to `text/plain`.
     */
    getEmailBody(type = "text/plain") {
        var _a, _b;
        return ((_b = (_a = this.getMediaByCategories(["body"])) === null || _a === void 0 ? void 0 : _a.filter((m) => m.contentType == type).shift()) !== null && _b !== void 0 ? _b : null);
    }
    /**
     * Get a media descriptor for an email history attachment of a provided type.
     * Allowed body types are returned in the Conversation.limits().emailHistoriesAllowedContentTypes array.
     * @param type Type of email history to request, defaults to `text/plain`.
     */
    getEmailHistory(type = "text/plain") {
        var _a, _b;
        return ((_b = (_a = this.getMediaByCategories(["history"])) === null || _a === void 0 ? void 0 : _a.filter((m) => m.contentType == type).shift()) !== null && _b !== void 0 ? _b : null);
    }
    _update(data) {
        var _a, _b, _c, _d, _e, _f;
        const updateReasons = [];
        if ((data.text || typeof data.text === "string") &&
            data.text !== this.state.body) {
            this.state.body = data.text;
            updateReasons.push("body");
        }
        if (data.subject && data.subject !== this.state.subject) {
            this.state.subject = data.subject;
            updateReasons.push("subject");
        }
        if (data.lastUpdatedBy && data.lastUpdatedBy !== this.state.lastUpdatedBy) {
            this.state.lastUpdatedBy = data.lastUpdatedBy;
            updateReasons.push("lastUpdatedBy");
        }
        if (data.author && data.author !== this.state.author) {
            this.state.author = data.author;
            updateReasons.push("author");
        }
        if (data.dateUpdated &&
            new Date(data.dateUpdated).getTime() !==
                (this.state.dateUpdated && this.state.dateUpdated.getTime())) {
            this.state.dateUpdated = new Date(data.dateUpdated);
            updateReasons.push("dateUpdated");
        }
        if (data.timestamp &&
            new Date(data.timestamp).getTime() !==
                (this.state.timestamp && this.state.timestamp.getTime())) {
            this.state.timestamp = new Date(data.timestamp);
            updateReasons.push("dateCreated");
        }
        const updatedAttributes = parseAttributes(data.attributes, `Got malformed attributes for the message ${this.sid}`, log);
        if (!isEqual(this.state.attributes, updatedAttributes)) {
            this.state.attributes = updatedAttributes;
            updateReasons.push("attributes");
        }
        const updatedAggregatedDelivery = data.delivery;
        const currentAggregatedDelivery = this.state.aggregatedDeliveryReceipt;
        const isUpdatedAggregateDeliveryValid = !!updatedAggregatedDelivery &&
            !!updatedAggregatedDelivery.total &&
            !!updatedAggregatedDelivery.delivered &&
            !!updatedAggregatedDelivery.failed &&
            !!updatedAggregatedDelivery.read &&
            !!updatedAggregatedDelivery.sent &&
            !!updatedAggregatedDelivery.undelivered;
        if (isUpdatedAggregateDeliveryValid) {
            if (!currentAggregatedDelivery) {
                this.state.aggregatedDeliveryReceipt = new AggregatedDeliveryReceipt(updatedAggregatedDelivery);
                updateReasons.push("deliveryReceipt");
            }
            else if (!currentAggregatedDelivery._isEquals(updatedAggregatedDelivery)) {
                currentAggregatedDelivery._update(updatedAggregatedDelivery);
                updateReasons.push("deliveryReceipt");
            }
        }
        // @fixme: This is a fairly heavy check.
        // Map data back and forth to perform a necessary field massage.
        const input = ((_a = data.medias) !== null && _a !== void 0 ? _a : [])
            .map((value) => { var _a; return (_a = __classPrivateFieldGet(this, _Message_instances, "m", _Message_wrapMedia).call(this, value)) === null || _a === void 0 ? void 0 : _a._state(); })
            .filter((el) => !(el === null));
        const current = ((_b = this.state.medias) !== null && _b !== void 0 ? _b : [])
            .map((value) => value._state())
            .filter((el) => !(el === null));
        if (!isEqual(input, current) ||
            !isEqual((_c = __classPrivateFieldGet(this, _Message_instances, "m", _Message_wrapMedia).call(this, data.media)) === null || _c === void 0 ? void 0 : _c._state(), (_d = this.state.media) === null || _d === void 0 ? void 0 : _d._state())) {
            __classPrivateFieldGet(this, _Message_instances, "m", _Message_assignMedias).call(this, (_e = data.medias) !== null && _e !== void 0 ? _e : null, (_f = data.media) !== null && _f !== void 0 ? _f : null);
            updateReasons.push("media");
        }
        if (updateReasons.length > 0) {
            this.emit("updated", { message: this, updateReasons: updateReasons });
        }
    }
    /**
     * Get the participant who is the author of the message.
     */
    async getParticipant() {
        let participant = null;
        if (this.state.participantSid) {
            participant = await this.conversation
                .getParticipantBySid(this.state.participantSid)
                .catch(() => {
                log.debug(`Participant with sid "${this.participantSid}" not found for message ${this.sid}`);
                return null;
            });
        }
        if (!participant && this.state.author) {
            participant = await this.conversation
                .getParticipantByIdentity(this.state.author)
                .catch(() => {
                log.debug(`Participant with identity "${this.author}" not found for message ${this.sid}`);
                return null;
            });
        }
        if (participant) {
            return participant;
        }
        let errorMessage = "Participant with ";
        if (this.state.participantSid) {
            errorMessage += "SID '" + this.state.participantSid + "' ";
        }
        if (this.state.author) {
            if (this.state.participantSid) {
                errorMessage += "or ";
            }
            errorMessage += "identity '" + this.state.author + "' ";
        }
        if (errorMessage === "Participant with ") {
            errorMessage = "Participant ";
        }
        errorMessage += "was not found";
        throw new Error(errorMessage);
    }
    /**
     * Get the delivery receipts of the message.
     */
    async getDetailedDeliveryReceipts() {
        let paginator = await this._getDetailedDeliveryReceiptsPaginator();
        let detailedDeliveryReceipts = paginator.items;
        while (paginator.hasNextPage) {
            paginator = await paginator.nextPage();
            detailedDeliveryReceipts = [
                ...detailedDeliveryReceipts,
                ...paginator.items,
            ];
        }
        return detailedDeliveryReceipts;
    }
    /**
     * Remove the message.
     */
    async remove() {
        await this.services.commandExecutor.mutateResource("delete", this.links.self);
        return this;
    }
    /**
     * Edit the message body.
     * @param body New body of the message.
     */
    async updateBody(body) {
        await this.services.commandExecutor.mutateResource("post", this.links.self, {
            body,
        });
        return this;
    }
    /**
     * Edit the message attributes.
     * @param attributes New attributes.
     */
    async updateAttributes(attributes) {
        await this.services.commandExecutor.mutateResource("post", this.links.self, {
            attributes: typeof attributes !== "undefined"
                ? JSON.stringify(attributes)
                : undefined,
        });
        return this;
    }
    /**
     * @deprecated
     * Get content URLs for all media attachments in the given set using a single operation.
     * @param contentSet Set of media attachments to query content URLs.
     */
    async attachTemporaryUrlsFor(contentSet) {
        // We ignore existing mcsMedia members of each of the media entries.
        // Instead we just collect their sids and pull new descriptors from a mediaSet GET endpoint.
        const sids = contentSet === null || contentSet === void 0 ? void 0 : contentSet.map((m) => m.sid);
        if (this.services.mcsClient && sids) {
            return (await this.services.mcsClient.mediaSetGet(sids)).map((item) => {
                return new Media(item, this.services);
            });
        }
        else {
            throw new Error("Media Content Service is unavailable");
        }
    }
    /**
     * Get content URLs for all media attachments in the given set using a single operation.
     * @param contentSet Set of media attachments to query content URLs.
     */
    getTemporaryContentUrlsForMedia(contentSet) {
        // We ignore existing mcsMedia members of each of the media entries.
        // Instead we just collect their sids and pull new descriptors from a mediaSet GET endpoint.
        const sids = contentSet.map((m) => m.sid);
        return this.getTemporaryContentUrlsForMediaSids(sids);
    }
    /**
     * Get content URLs for all media attachments in the given set of media sids using a single operation.
     * @param mediaSids Set of media sids to query for the content URL.
     */
    getTemporaryContentUrlsForMediaSids(mediaSids) {
        return new CancellablePromise(async (resolve, reject, onCancel) => {
            const mediaGetRequest = this.services.mcsClient.mediaSetGetContentUrls(mediaSids !== null && mediaSids !== void 0 ? mediaSids : []);
            if (!this.services.mcsClient || !mediaSids) {
                reject(new Error("Media Content Service is unavailable"));
                return;
            }
            onCancel(() => {
                mediaGetRequest.cancel();
            });
            try {
                const urls = await mediaGetRequest;
                resolve(urls);
            }
            catch (e) {
                reject(e);
            }
        });
    }
    /**
     * Get content URLs for all media attached to the message.
     */
    getTemporaryContentUrlsForAttachedMedia() {
        var _a;
        const media = this.attachedMedia;
        const sids = (_a = media === null || media === void 0 ? void 0 : media.map((m) => m.sid)) !== null && _a !== void 0 ? _a : [];
        return this.getTemporaryContentUrlsForMediaSids(sids);
    }
    async _getDetailedDeliveryReceiptsPaginator(options) {
        const messagesReceiptsUrl = this.configuration.links.messagesReceipts
            .replace("%s", this.conversation.sid)
            .replace("%s", this.sid);
        const url = new UriBuilder(messagesReceiptsUrl)
            .arg("PageToken", options === null || options === void 0 ? void 0 : options.pageToken)
            .arg("PageSize", options === null || options === void 0 ? void 0 : options.pageSize)
            .build();
        const response = await this.services.network.get(url);
        return new RestPaginator(response.body.delivery_receipts.map((x) => new DetailedDeliveryReceipt(x)), (pageToken, pageSize) => this._getDetailedDeliveryReceiptsPaginator({ pageToken, pageSize }), response.body.meta.previous_token, response.body.meta.next_token);
    }
    /**
     * Get the {@link ContentData} for this message. Resolves to `null` when
     * {@link Message.contentSid} is null.
     */
    getContentData() {
        return new CancellablePromise(async (resolve, reject, onCancel) => {
            if (this.state.contentSid === null) {
                resolve(null);
                return;
            }
            const bodies = this.getMediaByCategories(["body"]);
            if (bodies === null) {
                resolve(null);
                return;
            }
            const twilioPrefix = "application/x-vnd.com.twilio.rich.";
            const filteredMedias = bodies.filter((media) => media.contentType.startsWith(twilioPrefix));
            if (filteredMedias.length === 0) {
                resolve(null);
                return;
            }
            const contentMedia = filteredMedias[0];
            const urlPromise = contentMedia.getContentTemporaryUrl();
            onCancel(() => {
                urlPromise.cancel();
            });
            let url;
            try {
                url = await urlPromise;
            }
            catch (e) {
                reject(e);
                return;
            }
            if (url === null) {
                resolve(null);
                return;
            }
            const jsonStringPromise = new Promise((resolve, reject) => {
                let isCancelled = false;
                const xhr = new XHR();
                xhr.open("GET", url !== null && url !== void 0 ? url : "", true);
                xhr.responseType = "text";
                xhr.onreadystatechange = () => {
                    if (xhr.readyState !== 4 || isCancelled) {
                        return;
                    }
                    resolve(xhr.responseText);
                };
                xhr.onerror = () => {
                    reject(xhr.statusText);
                };
                onCancel(() => {
                    isCancelled = true;
                    xhr.abort();
                    reject(new Error("XHR has been aborted"));
                });
                xhr.send();
            });
            let json;
            try {
                const jsonString = await jsonStringPromise;
                json = JSON.parse(jsonString);
            }
            catch (e) {
                reject(e);
                return;
            }
            const dataType = contentMedia.contentType
                .replace(twilioPrefix, "")
                .replace(".", "/");
            resolve(parseVariant(dataType, json.data));
        });
    }
    /**
     * Get the {@link ChannelMetadata} for this message. Resolves to `null` if
     * the message doesn't have any channel metadata.
     */
    async getChannelMetadata() {
        if (!this.state.hasChannelMetadata) {
            return null;
        }
        return await this.services.channelMetadataClient.getChannelMetadata(this.conversation.sid, this.sid);
    }
    /**
     * Get recipients of the message.
     */
    async getMessageRecipients() {
        return await this.services.messageRecipientsClient.getRecipientsFromMessage(this.conversation.sid, this.sid);
    }
}
_Message_instances = new WeakSet(), _Message_wrapMedia = function _Message_wrapMedia(media, category = null) {
    return media
        ? category
            ? new Media(Object.assign(Object.assign({}, media), { category }), this.services)
            : new Media(media, this.services)
        : null;
}, _Message_assignMedias = function _Message_assignMedias(medias, media) {
    this.state.media = __classPrivateFieldGet(this, _Message_instances, "m", _Message_wrapMedia).call(this, media);
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    this.state.medias = medias
        ? medias.map((m) => __classPrivateFieldGet(this, _Message_instances, "m", _Message_wrapMedia).call(this, m)).filter((el) => el !== null)
        : media && !medias
            ? [__classPrivateFieldGet(this, _Message_instances, "m", _Message_wrapMedia).call(this, media, "media")].filter((el) => el !== null)
            : null;
};
/**
 * Fired when the properties or the body of the message has been updated.
 *
 * Parameters:
 * 1. object `data` - info object provided with the event. It has the following properties:
 *     * {@link Message} message - the message in question
 *     * {@link MessageUpdateReason}[] updateReasons - array of reasons for the update
 */
Message.updated = "updated";
__decorate([
    deprecated("getMediaByCategory", "getMediaByCategories"),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", Array)
], Message.prototype, "getMediaByCategory", null);
__decorate([
    validateTypes([nonEmptyString, "undefined"]),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Media)
], Message.prototype, "getEmailBody", null);
__decorate([
    validateTypes([nonEmptyString, "undefined"]),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Media)
], Message.prototype, "getEmailHistory", null);
__decorate([
    validateTypesAsync("string"),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", Promise)
], Message.prototype, "updateBody", null);
__decorate([
    validateTypesAsync(json),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object]),
    __metadata("design:returntype", Promise)
], Message.prototype, "updateAttributes", null);
__decorate([
    deprecated("attachTemporaryUrlsFor", "getTemporaryContentUrlsForMedia"),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", Promise)
], Message.prototype, "attachTemporaryUrlsFor", null);
__decorate([
    validateTypesAsync(nonEmptyArray("media", Media)),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", CancellablePromise)
], Message.prototype, "getTemporaryContentUrlsForMedia", null);
__decorate([
    validateTypesAsync(nonEmptyArray("strings", "string")),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Array]),
    __metadata("design:returntype", CancellablePromise)
], Message.prototype, "getTemporaryContentUrlsForMediaSids", null);

export { Message };
//# sourceMappingURL=message.js.map
